@@ -2,35 +2,83 @@ import logging
 
 logger = logging.getLogger("minimalKB." + __name__)
 
+import time
+
 from queue import Queue, Empty
-import json
 import traceback
 
-DEFAULT_MODEL = "default"
-
-import shlex
-from multiprocessing import Process
-
-hasRDFlib = False
 try:
     import rdflib
-    import rdflib.namespace
 
-    hasRDFlib = True
+    if rdflib.__version__ < "6.0.0":
+        logger.error("RDFlib >= 6.0.0 is required. Please upgrade.")
+        import sys
+
+        sys.exit(1)
 except ImportError:
-    logger.warn("RDFlib not available. You won't be able to load existing ontologies.")
-    pass
+    logger.error("RDFlib is required. Please install it.")
+    import sys
+
+    sys.exit(1)
+
+from rdflib import Graph, Dataset
+from rdflib.term import Literal, URIRef, Variable
+from rdflib.namespace import Namespace, RDF, RDFS
+
+has_reasoner = False
+try:
+    import reasonable
+
+    has_reasoner = True
+except ImportError:
+    logger.warning(
+        "reasonable OWL2 RL reasoner not available. Install it with `pip install reasonable`. Running without reasoning."
+    )
+
+
+import hashlib
+
+
+def stable_hash(s, p="", o="", model=""):
+    return hashlib.md5(("%s%s%s%s" % (s, p, o, model)).encode("utf-8")).hexdigest()
+
+
+DEFAULT_MODEL = "default"
+REASONER_RATE = 5  # Hz
+
+
+IRIS = {
+    "oro": "http://kb.openrobots.org#",
+    "cyc": "http://sw.opencyc.org/concept/",
+    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
+    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
+    "owl": "http://www.w3.org/2002/07/owl#",
+    "xsd": "http://www.w3.org/2001/XMLSchema#",
+}
+
+DEFAULT_PREFIX = "oro"
+
+default_ns = Namespace(IRIS[DEFAULT_PREFIX])
+
+SPARQL_PREFIXES = "".join(["PREFIX %s: <%s>\n" % (p, iri) for p, iri in IRIS.items()])
+SPARQL_PREFIXES += "PREFIX : <%s>\n" % IRIS[DEFAULT_PREFIX]
+SPARQL_PREFIXES += "BASE <%s>\n" % IRIS[DEFAULT_PREFIX]
+
+N3_PROLOGUE = " ".join(["@prefix %s: <%s>." % (p, iri) for p, iri in IRIS.items()])
+N3_PROLOGUE += "@prefix : <%s>. " % IRIS[DEFAULT_PREFIX]
+N3_PROLOGUE += "@base <%s>. " % IRIS[DEFAULT_PREFIX]
+N3_PROLOGUE += (
+    "@keywords a,true,false. "  # see https://www.w3.org/TeamSubmission/n3/#keywords
+)
+
 
 from .exceptions import KbServerError
 from minimalkb import __version__
 
-from .backends.sqlite import SQLStore
-
-# from backends.rdflib_backend import RDFlibStore
-
-from .services.simple_rdfs_reasoner import start_reasoner, stop_reasoner
 from .services import lifespan
 
+from .helpers import memoize
+
 
 def api(fn):
     fn._api = True
@@ -42,16 +90,85 @@ def compat(fn):
     return fn
 
 
+# @memoize
+def parse_stmts_to_graph(stmts):
+
+    data = N3_PROLOGUE
+
+    for stmt in stmts:
+        data += " %s . " % stmt
+
+    try:
+        return Graph().parse(data=data, format="n3")
+    except rdflib.plugins.parsers.notation3.BadSyntax:
+        raise KbServerError("invalid syntax for statements %s" % stmts)
+
+
+# @memoize
+def parse_stmts(stmts):
+
+    return list(parse_stmts_to_graph(stmts))
+
+
+# @memoize
 def parse_stmt(stmt):
 
-    tokens = stmt.split()
-    if len(tokens) < 3:
-        logger.error(
-            "Error while parsing the statement: %s. Only 2 tokens found" % stmt
-        )
-        raise RuntimeError("Malformed statement <%s>" % stmt)
+    try:
+        return list(Graph().parse(data=N3_PROLOGUE + "%s ." % stmt, format="n3"))[0]
+    except rdflib.plugins.parsers.notation3.BadSyntax:
+        raise KbServerError("invalid syntax for statement <%s>" % stmt)
 
-    return tokens[0], tokens[1], " ".join(tokens[2:])
+
+# @memoize
+def parse_term(term):
+
+    try:
+        # TODO: correct, but not super effective!
+        return list(
+            Graph().parse(data=N3_PROLOGUE + " <s> <p> %s ." % term, format="n3")
+        )[0][2]
+    except rdflib.plugins.parsers.notation3.BadSyntax:
+        raise KbServerError("invalid syntax for term <%s>" % term)
+
+
+def shorten_term(graph, term):
+
+    if isinstance(term, URIRef):
+        return graph.qname(term)
+    else:
+        return term.n3()
+
+
+def shorten(graph, stmt):
+
+    res = []
+    for t in stmt:
+        if isinstance(t, URIRef):
+            res.append(graph.qname(t))
+        else:
+            res.append(t.n3())
+    return res
+
+
+def shortenN(graph, stmts):
+    return [shorten(graph, s) for s in stmts]
+
+
+def shorten_graph(graph):
+    """returns a list of s,p,o statements contained in a graph, with their URIs shorten (eg, using prefixes when possible)"""
+    return [shorten(graph, s) for s in graph.triples([None, None, None])]
+
+
+def get_variables(stmt):
+    return [v for v in stmt if isinstance(v, Variable)]
+
+
+class dotdict(dict):
+    """dot.notation access to dictionary attributes"""
+
+    __getattr__ = dict.get
+    __setattr__ = dict.__setitem__
+    __delattr__ = dict.__delitem__
 
 
 class Event:
@@ -67,15 +184,11 @@ class Event:
         if type == Event.NEW_CLASS_INSTANCE:
             self.type = Event.NEW_INSTANCE
             self.var = "?instance"
-            self.patterns = [
-                parse_stmt("?instance rdf:type %s" % klass[0]) for klass in patterns
-            ]
+            self.patterns = ["?instance rdf:type %s" % klass[0] for klass in patterns]
         elif type == Event.NEW_CLASS_INSTANCE_ONE_SHOT:
             self.type = Event.NEW_INSTANCE_ONE_SHOT
             self.var = "?instance"
-            self.patterns = [
-                parse_stmt("?instance rdf:type %s" % klass[0]) for klass in patterns
-            ]
+            self.patterns = ["?instance rdf:type %s" % klass[0] for klass in patterns]
         else:
             self.type = type
             self.var = var
@@ -85,7 +198,7 @@ class Event:
         self.models = models
 
         self.id = "evt_" + str(
-            hash(
+            stable_hash(
                 self.type
                 + self.trigger
                 + (self.var if self.var else "")
@@ -100,9 +213,8 @@ class Event:
 
         self.previous_instances = set()
         if type in [Event.NEW_INSTANCE, Event.NEW_INSTANCE_ONE_SHOT]:
-            instances = self.kb.store.query(
-                [self.var], self.patterns, frozenset(self.models)
-            )
+
+            instances = self.kb.find([self.var], self.patterns, frozenset(self.models))
             logger.debug(
                 "Creating a NEW_INSTANCE event with initial instances %s" % instances
             )
@@ -120,7 +232,7 @@ class Event:
 
         if self.type in [Event.NEW_INSTANCE, Event.NEW_INSTANCE_ONE_SHOT]:
             instances = set(
-                self.kb.store.query([self.var], self.patterns, frozenset(self.models))
+                self.kb.find([self.var], self.patterns, frozenset(self.models))
             )
             newinstances = instances - self.previous_instances
 
@@ -143,7 +255,12 @@ class MinimalKB:
     MEMORYPROFILE_DEFAULT = ""
     MEMORYPROFILE_SHORTTERM = "SHORTTERM"
 
-    def __init__(self, filenames=None):
+    def __init__(self, filenames=None, enable_reasoner=True):
+
+        self.reasoner_enabled = enable_reasoner
+        if not self.reasoner_enabled:
+            logger.warn("Running without OWL2 RL reasoner.")
+
         _api = [
             getattr(self, fn) for fn in dir(self) if hasattr(getattr(self, fn), "_api")
         ]
@@ -151,11 +268,6 @@ class MinimalKB:
 
         self._api = {fn.__name__ + str(inspect.signature(fn)): fn for fn in _api}
 
-        self.store = SQLStore()
-        # self.store = RDFlibStore()
-
-        self.models = {DEFAULT_MODEL}
-
         apilist = [
             key + (" (compatibility)" if hasattr(val, "_compat") else "")
             for key, val in self._api.items()
@@ -172,7 +284,12 @@ class MinimalKB:
         self.active_evts = set()
         self.eventsubscriptions = {}
 
-        self.initialize_model(DEFAULT_MODEL)
+        # create a RDFlib dataset
+        self.ds = Dataset()
+        self.models = {}
+        self.create_model(DEFAULT_MODEL)
+
+        self._functionalproperties = frozenset()
 
         self.start_services()
 
@@ -184,114 +301,59 @@ class MinimalKB:
     def hello(self):
         return "MinimalKB, v.%s" % __version__
 
-    @compat
-    @api
-    def stats(self):
-        return {"version": __version__}
-
     @api
     def load(self, filename, models=None):
 
         models = self.normalize_models(models)
-        if hasRDFlib and (filename.endswith("owl") or filename.endswith("rdf")):
-            logger.info("Trying to load RDF file %s..." % filename)
-            g = rdflib.Graph()
-            nsm = rdflib.namespace.NamespaceManager(g)
-            # namespace_manager.bind(DEFAULT_NAMESPACE[0], self.default_ns)
-            g.parse(filename)
-            triples = []
-            for s, p, o in g:
-
-                # skip blank nodes
-                if (
-                    isinstance(s, rdflib.term.BNode)
-                    or isinstance(p, rdflib.term.BNode)
-                    or isinstance(o, rdflib.term.BNode)
-                ):
-                    continue
-                try:
-                    s = nsm.qname(s)
-                except:
-                    pass
-                try:
-                    p = nsm.qname(p)
-                except:
-                    pass
-                try:
-                    if isinstance(o, rdflib.term.Literal):
-                        o = o.toPython()
-                    else:
-                        o = nsm.qname(o)
-                except:
-                    pass
-                triples += [(s, p, o)]
-
-            logger.debug("Importing:\n%s" % triples)
-            for model in models:
-                self.store.add(triples, model)
-        else:
-            if filename.endswith("owl") or filename.endswith("rdf"):
-                logger.error(
-                    "Trying to load a RDF file, but RDFlib is not available"
-                    "Install first RDFlib for Python. Ignoring the file for "
-                    "now."
-                )
-                return
-
-            logger.info("Trying to load raw triples from %s..." % filename)
-            with open(filename, "r") as triples:
-                for model in models:
-                    self.store.add(
-                        [shlex.split(s.strip()) for s in triples.readlines()], models
-                    )
-
-    @compat
-    @api
-    def listAgents(self):
-        return list(self.models)
+        for model in models:
+            logger.info("Loading file <%s> in model <%s>" % (filename, model))
+            self.models[model].graph.parse(filename, publicID=IRIS[DEFAULT_PREFIX])
 
     @api
     def clear(self):
         logger.warn("Clearing the knowledge base!")
-        self.store.clear()
         self.active_evts.clear()
 
-        models = set(DEFAULT_MODEL)
-        self.initialize_model(DEFAULT_MODEL)
+        for m, g in self.models.items():
+            self.ds.remove_graph(g.graph)
 
-    @compat
-    @api
-    def reset(self):
-        self.clear()
-
-    @compat
-    @api
-    def listSimpleMethods(self):
-        return self.methods()
+        self.create_model(DEFAULT_MODEL)
 
     @api
     def methods(self):
         return list(self._api.keys())
 
     @api
-    def about(self, resource, models=None):
-        return self.store.about(resource, self.normalize_models(models))
+    def about(self, term, models=None):
 
-    @compat
-    @api
-    def lookupForAgent(self, agent, resource):
-        return self.lookup(resource, models=[agent])
+        result = []
+
+        models = self.normalize_models(models)
+
+        term = parse_term(term)
+
+        for model in models:
+            g = self.models[model].materialized_graph
+            for triple in g.triples((term, None, None)):
+                result.append(shorten(g, triple))
+            for triple in g.triples((None, term, None)):
+                result.append(shorten(g, triple))
+            for triple in g.triples((None, None, term)):
+                result.append(shorten(g, triple))
+
+        return result
 
     @api
     def lookup(self, resource, models=None):
+        models = self.normalize_models(models)
         logger.info(
             "Lookup for "
             + str(resource)
             + " in "
             + (str(models) if models else "default model.")
         )
-        models = self.normalize_models(models)
-        about = self.store.about(resource, models)
+        about = self.about(resource, models)
+
         if not about:
             logger.info("'%s' not found." % str(resource))
             return []
@@ -301,16 +363,128 @@ class MinimalKB:
         for s, p, o in about:
             if s == resource or p == resource or o == resource:
                 matching_concepts.add(resource)
-            if resource in o and p == "rdfs:label":  # resource is the label -> add s
-                matching_concepts.add(s)
 
-        res = [
-            (concept, self.store.typeof(concept, models))
-            for concept in matching_concepts
-        ]
+        res = [(concept, self.typeof(concept, models)) for concept in matching_concepts]
         logger.info("Found: " + str(res))
         return res
 
+    @api
+    def _instancesof(self, term, direct=False, models=[]):
+
+        if direct:
+            logger.warn(
+                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
+            )
+
+        models = self.normalize_models(models)
+        result = []
+
+        res = parse_term(term)
+        for model in models:
+            result += self.models[model].materialized_graph.subjects(RDF.type, res)
+
+        return result
+
+    @api
+    def classesof(self, term, direct=False, models=[]):
+        cls = self._classesof(term, direct, models)
+        return [
+            shorten_term(self.models[DEFAULT_MODEL].materialized_graph, c) for c in cls
+        ]
+
+    def _classesof(self, term, direct=False, models=[]):
+
+        if direct:
+            logger.warn(
+                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
+            )
+
+        models = self.normalize_models(models)
+        result = []
+
+        res = parse_term(term)
+        for model in models:
+            result += self.models[model].materialized_graph.objects(res, RDF.type)
+
+        return result
+
+    @api
+    def _subclassesof(self, term, direct=False, models=[]):
+
+        if direct:
+            logger.warn(
+                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
+            )
+
+        models = self.normalize_models(models)
+        result = []
+
+        res = parse_term(term)
+        for model in models:
+            result += self.models[model].materialized_graph.subjects(
+                RDFS.subClassOf, res
+            )
+
+        return result
+
+    @api
+    def _superclassesof(self, term, direct=False, models=[]):
+
+        if direct:
+            logger.warn(
+                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
+            )
+
+        models = self.normalize_models(models)
+        result = []
+
+        res = parse_term(term)
+        for model in models:
+            result += self.models[model].materialized_graph.objects(
+                res, RDFS.subClassOf
+            )
+
+        return result
+
+    def typeof(self, raw_term, models):
+
+        term = parse_term(raw_term)
+
+        if isinstance(term, Literal):
+            return "literal"
+
+        classes = self._classesof(raw_term, False, models)
+        if classes:
+            if "owl:ObjectProperty" in classes:
+                return "object_property"
+            elif "owl:DatatypeProperty" in classes:
+                return "datatype_property"
+            elif "owl:Class" in classes:
+                return "class"
+            else:
+                return "instance"
+
+        if (
+            self._instancesof(raw_term, False, models)
+            or self._subclassesof(raw_term, False, models)
+            or self._superclassesof(raw_term, False, models)
+        ):
+            return "class"
+
+        for model in models:
+            stmts_if_predicate = list(
+                self.models[model].materialized_graph.triples([None, term, None])
+            )
+
+            if stmts_if_predicate:
+                if isinstance(stmts_if_predicate[0][2], Literal):
+                    return "datatype_property"
+                else:
+                    return "object_property"
+
+        logger.warn("Concept <%s> has undecidable type." % term.n3())
+        return "undecided"
+
     @api
     def details(self, resource, models=None):
         """
@@ -343,7 +517,7 @@ class MinimalKB:
                     "id": "superClasses",
                     "values": [
                         {"id": r, "name": self.store.label(r, models)}
-                        for r in self.store.superclassesof(resource, True, models)
+                        for r in self._superclassesof(resource, True, models)
                     ],
                 }
             )
@@ -354,7 +528,7 @@ class MinimalKB:
                     "id": "subClasses",
                     "values": [
                         {"id": r, "name": self.store.label(r, models)}
-                        for r in self.store.subclassesof(resource, True, models)
+                        for r in self._subclassesof(resource, True, models)
                     ],
                 }
             )
@@ -365,7 +539,7 @@ class MinimalKB:
                     "id": "instances",
                     "values": [
                         {"id": r, "name": self.store.label(r, models)}
-                        for r in self.store.instancesof(resource, True, models)
+                        for r in self._instancesof(resource, True, models)
                     ],
                 }
             )
@@ -377,17 +551,12 @@ class MinimalKB:
                     "id": "classes",
                     "values": [
                         {"id": r, "name": self.store.label(r, models)}
-                        for r in self.store.classesof(resource, True, models)
+                        for r in self.store._classesof(resource, True, models)
                     ],
                 }
             ]
         return res
 
-    @compat
-    @api
-    def getResourceDetails(self, concept):
-        return self.details(concept)
-
     @api
     def check(self, stmts, models=None):
         logger.warn(
@@ -398,19 +567,36 @@ class MinimalKB:
         return self.exist(stmts, models)
 
     @api
-    def exist(self, stmts, models=None):
+    def exist(self, raw_stmts, models=None):
+        """Returns True if all the statements exist (eg, are materialised) in
+        all the given models.
+        """
+
+        models = self.normalize_models(models)
+
         logger.info(
             "Checking existence of "
-            + str(stmts)
+            + str(raw_stmts)
             + " in "
             + (str(models) if models else "default model.")
         )
-        stmts = [parse_stmt(s) for s in stmts]
+        stmts = [parse_stmt(s) for s in raw_stmts]
 
-        res = self.store.has(stmts, self.normalize_models(models))
+        vars = []
+        for stmt in stmts:
+            vars += [v.n3() for v in get_variables(stmt)]
 
-        logger.info("It exists" if res else "It does not exist.")
-        return res
+        # no variable? simply check if all the statements are asserted
+        if len(vars) == 0:
+            for model in models:
+                for stmt in stmts:
+                    if not stmt in self.models[model].materialized_graph:
+                        return False
+
+            return True
+
+        # else, run a query
+        return bool(self.find(vars, raw_stmts, models))
 
     @api
     def revise(self, stmts, policy):
@@ -425,7 +611,9 @@ class MinimalKB:
 
         if isinstance(stmts, str):
             raise KbServerError("A list of statements is expected")
-        stmts = [parse_stmt(s) for s in stmts]
+
+        subgraph = parse_stmts_to_graph(stmts)
+        parsed_stmts = "\n\t- ".join([" ".join(s) for s in shorten_graph(subgraph)])
 
         if type(policy) != dict:
             raise KbServerError("Expected a dictionary as policy")
@@ -439,21 +627,18 @@ class MinimalKB:
                 "Adding to "
                 + str(list(models))
                 + ":\n\t- "
-                + "\n\t- ".join([str(s) for s in stmts])
+                + parsed_stmts
                 + (" (lifespan: %ssec)" % lifespan if lifespan else "")
             )
             for model in models:
-                self.store.add(stmts, model, lifespan=lifespan)
+                self.models[model].graph += subgraph  # TODO: lifespan
+                self.models[model].is_dirty = True
 
         if policy["method"] == "retract":
-            logger.info(
-                "Deleting from "
-                + str(list(models))
-                + ":\n\t- "
-                + "\n\t- ".join([str(s) for s in stmts])
-            )
+            logger.info("Deleting from " + str(list(models)) + ":\n\t- " + parsed_stmts)
             for model in models:
-                self.store.delete(stmts, model)
+                self.models[model].graph -= subgraph
+                self.models[model].is_dirty = True
 
         if policy["method"] in ["update", "safe_update", "revision"]:
 
@@ -463,11 +648,17 @@ class MinimalKB:
                 "Updating "
                 + str(list(models))
                 + " with:\n\t- "
-                + "\n\t- ".join([str(s) for s in stmts])
+                + parsed_stmts
                 + (" (lifespan: %ssec)" % lifespan if lifespan else "")
             )
             for model in models:
-                self.store.update(stmts, model, lifespan=lifespan)
+                for s, p, o in subgraph.triples([None, None, None]):
+
+                    if p in self._functionalproperties:
+                        self.models[model].graph.set((s, p, o))  # TODO: lifespan
+                    else:
+                        self.models[model].graph.add((s, p, o))  # TODO: lifespan
+                self.models[model].is_dirty = True
 
         self.onupdate()
 
@@ -477,29 +668,9 @@ class MinimalKB:
             stmts, {"method": "add", "models": models, "lifespan": lifespan}
         )
 
-    @compat
-    @api
-    def safeAdd(self, stmts, lifespan=0):
-        return self.revise(stmts, {"method": "safe_add", "lifespan": lifespan})
-
-    @compat
-    @api
-    def addForAgent(self, agent, stmts, lifespan=0):
-        return self.add(stmts, [agent], lifespan)
-
-    @api
-    def retract(self, stmts, models=None):
-        return self.revise(stmts, {"method": "retract", "models": models})
-
-    @compat
     @api
     def remove(self, stmts, models=None):
-        return self.retract(stmts, models)
-
-    @compat
-    @api
-    def removeForAgent(self, agent, stmts):
-        return self.retract(stmts, [agent])
+        return self.revise(stmts, {"method": "retract", "models": models})
 
     @api
     def update(self, stmts, models=None, lifespan=0):
@@ -507,13 +678,54 @@ class MinimalKB:
             stmts, {"method": "update", "models": models, "lifespan": lifespan}
         )
 
-    @compat
     @api
-    def findForAgent(self, agent, var, stmts):
-        return self.find([var], stmts, None, [agent])
+    def sparql(self, query, model=None):
+        """
+        performs a raw SPARQL query on a given model ('default' by default).
+        The SPARQL PREFIX and BASE are automatically added, no need to do it
+        manually (even though you can if you want to use non-standard
+        prefixes).
+
+        Note that you are responsible for writing a syntactically corret SPARQL
+        query. In particualar, all non-literal/non-variable terms must have a
+        namespace (or a prefix).
+
+        Results is returned as a JSON object that follow the standard JSON
+        serialization of SPARQL Results
+        (https://www.w3.org/TR/2013/REC-sparql11-results-json-20130321/)
+
+        Example:
+        ```
+        >>> kb += ["myself age 40"]
+        >>> kb.sparql("SELECT ?age WHERE { :myself :age ?age . }")
+        {'results':
+            {'bindings': [
+                {'age':
+                    {'type': 'literal',
+                     'value': '40',
+                     'datatype': 'http://www.w3.org/2001/XMLSchema#integer'}
+                }]
+            },
+          'head': {'vars': ['age']}
+         }
+
+        """
+
+        models = self.normalize_models(model)
+        if len(models) != 1:
+            logger.error("sparql() can only be executed on a single model")
+            return None
+
+        model = list(models)[0]
+
+        sparql_res = self._sparql(model, query)
+
+        import json
+
+        return json.loads(sparql_res.serialize(format="json"))
 
     @api
-    def find(self, vars, patterns, constraints=None, models=None):
+    def find(self, vars, patterns, models=None):
         """
         Depending on the arguments, three differents
         behaviours are possible:
@@ -527,8 +739,6 @@ class MinimalKB:
             possible combination of values for the different variables. For
             instance, find(["?agent", "?action"], ["?agent desires ?action", "?action rdf:type Jump"])
             would return something like: [{"agent":"james", "action": "jumpHigh"}, {"agent": "laurel", "action":"jumpHigher"}]
-
-        Note that 'constraints' is currently not supported.
         """
 
         if not vars:
@@ -539,6 +749,11 @@ class MinimalKB:
         models = self.normalize_models(models)
 
         patterns = [parse_stmt(p) for p in patterns]
+        parsed_patterns = "\n\t- ".join(
+            [" ".join(s) for s in shortenN(self.models[DEFAULT_MODEL].graph, patterns)]
+        )
+
+        normalised_patterns = [" ".join([t.n3() for t in stmt]) for stmt in patterns]
 
         logger.info(
             "Searching "
@@ -546,29 +761,50 @@ class MinimalKB:
             + " in models "
             + str(models)
             + " matching:\n\t- "
-            + "\n\t- ".join([str(p) for p in patterns])
+            + parsed_patterns
         )
 
-        res = self.store.query(vars, patterns, models)
+        res = []
+
+        named_variables = self.named_variables(vars)
+        # if we have named variable + extra 'anonymous' variable (coming from
+        # eg '*'), ignore them
+        if named_variables and len(named_variables) != len(vars):
+            vars = named_variables
+
+        # remove the '?' at the start of the variables
+        vars_naked = [var[1:] for var in vars]
+
+        for model in models:
+            q = "SELECT %s WHERE {\n" % " ".join(vars)
+            for p in normalised_patterns:
+                q += "%s .\n" % p
+            q += "}"
+
+            sparql_res = self._sparql(model, q)
+
+            if len(named_variables) > 1:
+                res += [
+                    dict(zip(vars_naked, r))
+                    for r in [
+                        shorten(self.models[model].graph, row) for row in sparql_res
+                    ]
+                ]
+            else:
+                res += [shorten(self.models[model].graph, row) for row in sparql_res]
+
+        # if a single variable is requested, 'unpack' the result
+        if len(vars) == 1:
+            res = [r[0] for r in res]
 
         logger.info("Found: " + str(res))
         return res
 
-    @api
-    def findmpe(self, vars, pattern, constraints=None, models=None):
-        """Finds the most probable explanation. Strictly equivalent to
-        'find' until we support probabilities.
-        """
-        return find(self, vars, pattern, constraints=None, models=None)
-
     @api
     def subscribe(self, type, trigger, var, patterns, models=None):
 
         models = self.normalize_models(models)
 
-        if type not in [Event.NEW_CLASS_INSTANCE, Event.NEW_CLASS_INSTANCE_ONE_SHOT]:
-            patterns = [parse_stmt(p) for p in patterns]
-
         logger.info(
             "Registering a new event: %s %s for %s on %s"
             % (type, trigger, var, patterns)
@@ -582,38 +818,6 @@ class MinimalKB:
 
         return event.id
 
-    @compat
-    @api
-    def registerEvent(self, type, trigger, patterns):
-        return self.subscribe(type, trigger, None, patterns)
-
-    @compat
-    @api
-    def discriminateForAgent(self, *args):
-        raise NotImplementedError("discriminateForAgent not implemented in MinimalKB")
-
-    @compat
-    @api
-    def getLabel(self, concept):
-        return self.store.label(concept)
-
-    @compat
-    @api
-    def getDirectClassesOf(self, concept):
-        return self.getClassesOf(concept, True)
-
-    @compat
-    @api
-    def getClassesOf(self, concept, direct=False):
-        classes = self.classesof(concept, direct)
-
-        return {cls: self.getLabel(cls) for cls in classes}
-
-    @api
-    def classesof(self, concept, direct=False, models=None):
-        models = self.normalize_models(models)
-        return self.store.classesof(concept, direct, models)
-
     @api
     def close(self):
         """This code should actually never be called: the 'close' request
@@ -622,9 +826,42 @@ class MinimalKB:
         """
         raise NotImplementedError
 
+    @api
+    def debugger(self):
+        import pdb
+
+        pdb.set_trace()
+
     ################################################################################
     ################################################################################
+
+    def _sparql(self, model, query):
+        q = SPARQL_PREFIXES  # TODO: as a (potential?) optimization, pass initNs to graph.query, instead of adding the PREFIX strings to the query
+        q += query
+
+        logger.debug("Executing SPARQL query in model: %s\n%s" % (model, q))
+
+        return self.models[model].materialized_graph.query(q)
+
+    def named_variables(self, vars):
+        """
+        Returns the list of variable that have been explicitely named (eg,
+        not the 'stars' wildcards, that would have been replaced by anonymous
+        variable starting with '__' by pykb)
+        """
+
+        return [v for v in vars if not v.startswith("?__")]
+
     def onupdate(self):
+
+        # need to materialise as soon as possible after the model has been
+        # changed so that events are triggered in a timely fashion
+        self.materialise()
+
+        self._functionalproperties = frozenset(
+            self._instancesof("owl:FunctionalProperty", False)
+        )
+
         for e in self.active_evts:
             if e.evaluate():
                 clients = self.eventsubscriptions[e.id]
@@ -637,23 +874,66 @@ class MinimalKB:
                 if not e.valid:
                     self.active_evts.discard(e)
 
-    def start_services(self, *args):
-        self._reasoner = Process(target=start_reasoner, args=("kb.db",))
-        self._reasoner.start()
+    def materialise(self, models=None):
 
-        self._lifespan_manager = Process(target=lifespan.start_service, args=("kb.db",))
-        self._lifespan_manager.start()
+        start = time.time()
+
+        if models is None:
+            models = self.models.keys()
+
+        for model in models:
+            g = self.models[model]
+            if not g.is_dirty:
+                continue
+
+            if not has_reasoner or not self.reasoner_enabled:
+                g.materialized_graph = g.graph
+            else:
+                r = reasonable.PyReasoner()
+                r.from_graph(g.graph)
+
+                g.materialized_graph = Graph()
+                g.materialized_graph.namespace_manager = g.graph.namespace_manager
+                g.materialized_graph += r.reason()
+
+            g.is_dirty = False
+
+        end = time.time()
+
+        if end - start > 0.0001:  # did we actually classify anything?
+            logger.debug(
+                "Materialisation performed by reasoner in %.1fms"
+                % ((end - start) * 1000)
+            )
+
+    def start_services(self, *args):
+
+        # self._lifespan_manager = Process(
+        #    target=lifespan.start_service, args=("kb.db",)
+        # )
+        # self._lifespan_manager.start()
+        pass
 
     def stop_services(self):
-        self._reasoner.terminate()
-        self._lifespan_manager.terminate()
 
-        self._reasoner.join()
-        self._lifespan_manager.join()
+        # self._lifespan_manager.join()
+        pass
 
-    def initialize_model(self, model):
-        self.add(
-            ["owl:Thing rdf:type owl:Class", "owl:Nothing rdf:type owl:Class"], model
+    def create_model(self, model):
+        g = self.ds.graph(IRIS[DEFAULT_PREFIX] + model)
+
+        # configure our namespace prefixes
+        for p, iri in IRIS.items():
+            g.bind(p, iri)
+
+        g.bind("", IRIS[DEFAULT_PREFIX])
+
+        self.models[model] = dotdict(
+            {
+                "graph": g,
+                "is_dirty": True,  # stores whether models have changes that would require re-classification
+                "materialized_graph": Graph(),
+            }
         )
 
     def normalize_models(self, models):
@@ -673,8 +953,7 @@ class MinimalKB:
                 # do we have a new model? initialise it.
                 for model in models:
                     if model not in self.models:
-                        self.models.add(model)
-                        self.initialize_model(model)
+                        self.create_model(model)
 
                 return frozenset(models)
         else:
@@ -698,7 +977,7 @@ class MinimalKB:
             res = None
             if args or kwargs:
                 res = f(*args, **kwargs)
-                if name in ["subscribe", "registerEvent"]:
+                if name in ["subscribe"]:
                     self.eventsubscriptions.setdefault(res, []).append(client)
             else:
                 res = f()
